#include <vcl.h>
#pragma hdrstop
#include <tlhelp32.h>
#include <winperf.h>


#include "getmod.h"
#pragma package(smart_init)
//---------------------------------------------------------------------------


// MODULENAMEFROMHANDLE
// eg. AnsiString mod=ModuleNameFromHandle(hwnd);
// Windows NT and 95/98 use different code to get the current module name.
// For Win95 we use the ToolHelp functions from inside KERNEL32
// For NT there are two possibilities: either using PSAPI.DLL or do it
// yourself by querying the performance-data in the registry.
// But PSAPI.DLL only comes with the Platform SDK and some people (me!)
// might not have it. I've therefore preferred the second option.
// The Win95 ToolHelp functions return a full path name (c:\apps\bcb\bcb.exe)
// The NT registry ones return only the bare module name (BCB)
// I've therefore abbreviated win95 ones down as well to the lowest common
// denominator.
// The Win95 and NT/PSAPI code is from a newsgroup article by Bertrand Le Guern/
// The NT code comes from KB article Q119163
// I've modified them a bit to make them more concise and C++/VCL like

AnsiString __fastcall ModuleNameFromProcessId95(DWORD pid)
{ HINSTANCE hLib=LoadLibrary("KERNEL32.DLL"); if (hLib==NULL) return "";
  typedef HANDLE (WINAPI *PFN_SNAPSHOT)(DWORD dwFlags, DWORD th32ProcessID);
  typedef BOOL (WINAPI *PFN_PROCESS32WALK)(HANDLE hSnapshot, LPPROCESSENTRY32 lppe);
  typedef BOOL (WINAPI *PFN_MODULE32WALK)(HANDLE hSnapshot, LPMODULEENTRY32 lpme);
  PFN_SNAPSHOT pfnCreateToolhelp32Snapshot=(PFN_SNAPSHOT)GetProcAddress(hLib,"CreateToolhelp32Snapshot");
  PFN_PROCESS32WALK pfnProcess32First=(PFN_PROCESS32WALK)GetProcAddress(hLib,"Process32First");
  PFN_PROCESS32WALK pfnProcess32Next=(PFN_PROCESS32WALK)GetProcAddress(hLib,"Process32Next");
  if (pfnCreateToolhelp32Snapshot==NULL || pfnProcess32First==NULL || pfnProcess32Next==NULL) {FreeLibrary(hLib); return "";}
  // create a process snapshot and walk through the process list, adding
  // each process to the top listbox
  HANDLE hSnapshot = pfnCreateToolhelp32Snapshot(TH32CS_SNAPPROCESS,0);
  if (hSnapshot==NULL) {FreeLibrary(hLib);return "";}
  PROCESSENTRY32 pe32; ZeroMemory(&pe32,sizeof(pe32)); pe32.dwSize = sizeof(pe32);
  AnsiString s="";
  bool keepgoing=pfnProcess32First(hSnapshot,&pe32);
  while (keepgoing)
  { if (pe32.th32ProcessID==pid) {s=pe32.szExeFile;keepgoing=false;}
    else keepgoing=pfnProcess32Next(hSnapshot,&pe32);
  }
  CloseHandle(hSnapshot);
  FreeLibrary(hLib);
  s=s.UpperCase();
  return ChangeFileExt(ExtractFileName(s),"");
}



int __fastcall GetIndex(AnsiString title);
AnsiString __fastcall ModuleNameFromProcessIdNT(DWORD dwProcessId)
{ int ProcessIndex=GetIndex("Process"); if (ProcessIndex==-1) return "";
  int IdIndex=GetIndex( "ID Process"); if (IdIndex==-1) return "";
  // Get memory for PPERF_DATA_BLOCK, and get data. We do it in a loop for some reason
  // to get the right size. The initial 12000 is just a guess.
  DWORD dwBytes=12000; PPERF_DATA_BLOCK pdb = (PPERF_DATA_BLOCK) HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,dwBytes);
  while (RegQueryValueEx(HKEY_PERFORMANCE_DATA,(AnsiString(ProcessIndex)).c_str(),NULL,NULL,(LPBYTE)pdb,&dwBytes) == ERROR_MORE_DATA )
  { dwBytes += 1000;
    pdb = (PPERF_DATA_BLOCK) HeapReAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,(LPVOID)pdb,dwBytes);
  }
  PPERF_OBJECT_TYPE pot = (PPERF_OBJECT_TYPE)((PBYTE)pdb + pdb->HeaderLength);
  // Get the first counter definition.
  PPERF_COUNTER_DEFINITION  pcd = (PPERF_COUNTER_DEFINITION)((PBYTE)pot + pot->HeaderLength);
  DWORD dwProcessIdOffset;
  for (int i=0; i< (int)pot->NumCounters; i++ )
  { if (pcd->CounterNameTitleIndex == (DWORD)IdIndex)
    { dwProcessIdOffset = pcd->CounterOffset;
      break;
    }
    pcd = ((PPERF_COUNTER_DEFINITION)((PBYTE)pcd + pcd->ByteLength));
  }
  // Get the first instance of the object.
  PPERF_INSTANCE_DEFINITION pid = (PPERF_INSTANCE_DEFINITION)((PBYTE)pot + pot->DefinitionLength);
  // Get the name of the first process.
  PPERF_COUNTER_BLOCK pcb = (PPERF_COUNTER_BLOCK) ((PBYTE)pid + pid->ByteLength );
  DWORD CurrentProcessId = *((DWORD *) ((PBYTE)pcb + dwProcessIdOffset));
  // Find the process object for PID passed in, then print its filename.
  wchar_t *sres=L"";  BOOL bContinue = TRUE;
  for(int i = 1; i < pot->NumInstances && bContinue; i++ )
  { if( CurrentProcessId == dwProcessId )
    { sres=(wchar_t*) ((PBYTE)pid + pid->NameOffset) ;
      bContinue = FALSE;
    }
    else
    { pid = (PPERF_INSTANCE_DEFINITION) ((PBYTE)pcb + pcb->ByteLength);
      pcb = (PPERF_COUNTER_BLOCK) ((PBYTE)pid + pid->ByteLength);
      CurrentProcessId = *((DWORD *)((PBYTE)pcb + dwProcessIdOffset));
    }
  }
  AnsiString s=WideCharToString(sres);
  HeapFree(GetProcessHeap(), 0, (LPVOID)pdb) ;
  RegCloseKey( HKEY_PERFORMANCE_DATA );
  return ChangeFileExt(s.UpperCase(),"");
}
int __fastcall GetIndex(AnsiString title)
{ // We're going to get a buffer from this particular key.
  HANDLE hKeyIndex; LONG res=RegOpenKeyEx( HKEY_LOCAL_MACHINE,"SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Perflib\\009",0, KEY_READ,&hKeyIndex); if (res!=ERROR_SUCCESS) return -1;
  DWORD CountersSize; RegQueryValueEx(hKeyIndex,"Counters",NULL, NULL, NULL,&CountersSize );
  char *CountersBuf=new char[CountersSize]; ZeroMemory(CountersBuf,CountersSize);
  RegQueryValueEx(hKeyIndex,"Counters",NULL,NULL,(LPBYTE)CountersBuf,&CountersSize);
  // Find the index value for whatever.
  int index=-1;
  unsigned int i=0;
  while (i<CountersSize && index==-1)
  { AnsiString curindex=""; while (CountersBuf[i]!=0) {curindex=curindex+CountersBuf[i]; i++;}
    i++; AnsiString curobj="";
    while (CountersBuf[i]!=0) {curobj=curobj+CountersBuf[i]; i++;}
    i++; if (CountersBuf[i]==0) i++;
    if (curobj==title) index=curindex.ToInt();
  }
  delete[] CountersBuf;
  RegCloseKey(hKeyIndex);
  return index;
}



AnsiString __fastcall ModuleNameFromProcessIdNT2(DWORD pid)
{ HINSTANCE hLib=LoadLibrary( "PSAPI.DLL"); if (hLib==NULL) return "";
  typedef BOOL (WINAPI *PFN_ENUMPROC)(DWORD * lpidProcess,DWORD cb,DWORD *cbNeeded);
  typedef BOOL (WINAPI *PFN_ENUMMOD)( HANDLE hProcess,HMODULE *lphModule,DWORD cb,LPDWORD lpcbNeeded);
  typedef DWORD (WINAPI *PFN_GETMODNAME)( HANDLE hProcess,HMODULE hModule,LPSTR lpFilename,DWORD nSize);
  PFN_ENUMPROC pfnEnumProcesses=(PFN_ENUMPROC)GetProcAddress(hLib,"EnumProcesses");
  PFN_ENUMMOD pfnEnumProcessModules=(PFN_ENUMMOD)GetProcAddress(hLib,"EnumProcessModules");
  PFN_GETMODNAME pfnGetModuleBaseNameA=(PFN_GETMODNAME)GetProcAddress(hLib,"GetModuleBaseNameA");
  PFN_GETMODNAME pfnGetModuleFileNameExA=(PFN_GETMODNAME)GetProcAddress(hLib,"GetModuleFileNameExA");
  if (pfnEnumProcesses==NULL || pfnEnumProcessModules==NULL || pfnGetModuleBaseNameA==NULL || pfnGetModuleFileNameExA==NULL)
  { FreeLibrary(hLib); return "";
  }
  HANDLE hProcess = OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, FALSE, pid);
  if (hProcess==NULL) {FreeLibrary(hLib);return "";}
  AnsiString s="";
  HMODULE hMod; DWORD cbNeeded; bool isit=pfnEnumProcessModules(hProcess,&hMod,sizeof(hMod),&cbNeeded);
  if (isit)
  { char *c=new char[cbNeeded];
    pfnGetModuleFileNameExA(hProcess,hMod,c,cbNeeded);
    s=AnsiString(c);
    delete[] c;
  }
  CloseHandle( hProcess );
  FreeLibrary(hLib);
  return s;
}









AnsiString __fastcall ModuleNameFromProcessId(DWORD pid)
{ OSVERSIONINFO osvi; ZeroMemory(&osvi,sizeof(osvi)); osvi.dwOSVersionInfoSize=sizeof(OSVERSIONINFO);
  GetVersionEx(&osvi);
  if (osvi.dwPlatformId==VER_PLATFORM_WIN32_NT) return ModuleNameFromProcessIdNT(pid);
  else return ModuleNameFromProcessId95(pid);
}

AnsiString __fastcall ModuleNameFromHWnd(HWND hwnd)
{ DWORD pid; GetWindowThreadProcessId(hwnd,&pid);
  return ModuleNameFromProcessId(pid);
}

